//blocks.cc: Implementation file for blocks.h

#include "signal.h"
#include "cell.h"
#include "blocks.h"
#include <iostream.h>

///////////////////////////////////////////////////////////////////////////////
//Model of a Configurable Logic Block

CLB_class::CLB_class()
{
  Clear();
} //CLB constructor


CLB_class::~CLB_class()
{
} //CLB destructor


void CLB_class::Clear()
{
  //erase all state, initialize all values to unknown
  //set previous output to unknown

  current_output = Unknown;
  previous_output = Unknown;
  current_test_output = Unknown;
  previous_test_output = Unknown;
  Lookup_table.Clear();
  Mux0.Clear();
  Mux1.Clear();
  Mux2.Clear();
  Flop.Clear();
} //Clear()


void CLB_class::Update(CLB_input current_inputs)
{
  //Update configuration (if necessary)
  //propagate inputs to output (update state)
  //Since there is feedback in the CLB, make just one pass.
  //Make subsequent calls for convergence

  //First, update shift registers.
  Lookup_table.Update(current_inputs.clk_a, current_inputs.clk_b,
		      current_inputs.config_enable, 
		      current_inputs.lut_config_in); //LUT
  //0->2->1
  signal_type old_Mux0_shift_output = Mux0.Get_Shift_Output();
  signal_type old_Mux2_shift_output = Mux2.Get_Shift_Output();
  Mux0.Update(current_inputs.clk_a, current_inputs.clk_b,
	      current_inputs.config_enable, 
	      current_inputs.mux_config_in);
  Mux2.Update(current_inputs.clk_a, current_inputs.clk_b,
	      current_inputs.config_enable, 
	      old_Mux0_shift_output);
  Mux1.Update(current_inputs.clk_a, current_inputs.clk_b,
	      current_inputs.config_enable, 
	      old_Mux2_shift_output);

  //Shift registers updated, do scan mux and flop
  signal_type scan_mux_output;
  scan_mux_output = GMUX2(current_output, current_inputs.scan_in,
			  current_inputs.test_mode);
  Flop.Update(current_inputs.clk_a, current_inputs.clk_b,
	      current_inputs.reset_flop, scan_mux_output);

  //All flops updated, take a pass at the logic
  //Do each mux, then the lookup table.
  signal_type Mux0_inputs[] = {current_inputs.south,current_inputs.west,
			       current_inputs.north,current_inputs.east};
  signal_type Mux1_inputs[] = {current_inputs.south,Flop.Get_Output(),
			       current_inputs.north,current_output};
  signal_type Mux2_inputs[] = {current_inputs.south,current_inputs.west,
			       current_inputs.north,current_inputs.east};
  signal_type Mux_outputs[3];
  Mux_outputs[0] = Mux0.Get_Output(Mux0_inputs); 
  Mux_outputs[1] = Mux1.Get_Output(Mux1_inputs);
  Mux_outputs[2] = Mux2.Get_Output(Mux2_inputs);

  signal_type LUT_output = Lookup_table.Get_Output(Mux_outputs);
  
  //Now update state of block
  previous_output = current_output;
  previous_test_output = current_test_output;
  current_output = LUT_output;
  current_test_output = Flop.Get_Output();
} //Update()


bool CLB_class::Converged()
{
  if((current_output == previous_output)&&
     (current_test_output == previous_test_output))
    return true;
  else
    return false;
} //Converged()


signal_type CLB_class::Get_Output()
{
  //return output of cell
  signal_type output = current_output;
  return output;
} //Get_Output()


signal_type CLB_class::Get_Test_Output()
{
  //return output of cell
  signal_type output = current_test_output;
  return output;
} //Get_Test_Output()


signal_type CLB_class::Get_Lut_Config_Out()
{
  //output of first config string
  signal_type output = Lookup_table.Get_Shift_Output();
  return output;
} //Get_Lut_Config_Out()


signal_type CLB_class::Get_Mux_Config_Out()
{
  //output of second config string
  signal_type output = Mux1.Get_Shift_Output();
  return output;
} //Get_Mux_Config_Out()


signal_type CLB_class::Get_Lut_Config_Bit(unsigned int index)
{
 //get arbitrary bit
  return Lookup_table.Get_Config_Bit(index);
} //Get_Lut_Config_Bit


signal_type CLB_class::Get_Mux0_Config_Bit(unsigned int index)
{
 //get arbitrary bit
  return Mux0.Get_Config_Bit(index);
} //Get_Mux0_Config_Bit


signal_type CLB_class::Get_Mux1_Config_Bit(unsigned int index)
{
 //get arbitrary bit
  return Mux1.Get_Config_Bit(index);
} //Get_Mux1_Config_Bit


signal_type CLB_class::Get_Mux2_Config_Bit(unsigned int index)
{
 //get arbitrary bit
  return Mux2.Get_Config_Bit(index);
} //Get_Mux2_Config_Bit


///////////////////////////////////////////////////////////////////////////////
//Model of a Configurable Sensor Block

CSB_class::CSB_class()
{
  //constructor
  Clear();
}


CSB_class::~CSB_class()
{
  //destructor
}


void CSB_class::Clear()
{
  //erase all state, initialize all values to unknown
  current_output = Unknown;
  current_test_output = Unknown;
  previous_output = Unknown;
  previous_test_output = Unknown;
  Optical_Receiver.Clear();
  Mux.Clear();
  Flop.Clear();
} //Clear()


void CSB_class::Update(CSB_input inputs)
{
  //Update configuration (if necessary)
  //propagate inputs to output (update state)

  //First, get receiver output (a primary input)
  signal_type receiver_output = Optical_Receiver.Get_Output(inputs.sensor);

  //Next, evaluate optical-config mux to get input to config chain
  signal_type Optical_Config_Mux_Output = GMUX2(inputs.config_in,
						receiver_output, 
						inputs.config_opt);
  //Next, update shift registers.
  signal_type Old_Mux_Config_Output = Mux.Get_Shift_Output();
  Mux.Update(inputs.clk_a, inputs.clk_b,
	     inputs.config_enable, Optical_Config_Mux_Output);
  Optical_Receiver.Update(inputs.clk_a, inputs.clk_b,
			  inputs.config_enable, Old_Mux_Config_Output);

  //Shift registers updated, do scan mux and flop
  signal_type scan_mux_output;
  scan_mux_output = GMUX2(current_output, inputs.scan_in,
			  inputs.test_mode);
  Flop.Update(inputs.clk_a, inputs.clk_b,
	      inputs.reset_flop, scan_mux_output);

  //All flops updated, do logic
  signal_type Mux_inputs[] = {inputs.south, inputs.west, inputs.north, 
			      inputs.east, receiver_output, inputs.clk_a,
			      inputs.clk_b,
			      current_output //This is a flaw in the design
  };
  signal_type Mux_output = Mux.Get_Output(Mux_inputs);
  
  //Now update state of block
  previous_output = current_output;
  previous_test_output = current_test_output;
  current_output = Mux_output;
  current_test_output = Flop.Get_Output();
  
} //Update()


bool CSB_class::Converged()
{  
  if((current_output == previous_output)&&
     (current_test_output == previous_test_output))
    return true;
  else
    return false;
} //Converged()

signal_type CSB_class::Get_Output()
{
  //return output of cell
  return current_output;
} //Get_Output()


signal_type CSB_class::Get_Test_Output()
{
  return current_test_output;
} //Get_Test_Output()


signal_type CSB_class::Get_Config_Out()
{
  //return output of cell
  signal_type output = Optical_Receiver.Get_Shift_Output();
  return output;
} //Get_Config_Out()


signal_type CSB_class::Get_Receiver_Config_Bit(unsigned int index)
{
  signal_type output = Optical_Receiver.Get_Config_Bit(index);
  return output;
} //Get_Receiver_Config_Bit


signal_type CSB_class::Get_Mux_Config_Bit(unsigned int index)
{
  signal_type output = Mux.Get_Config_Bit(index);
  return output;
} //Get_Mux_Config_Bit


///////////////////////////////////////////////////////////////////////////////

//useful pin-related function
void Print_pin_type(pin_type type)
{
  switch(type)
    {
    case io:
      cout<<"io";
      break;
    case dedicated_input:
      cout<<"in";
      break;
    case dedicated_output:
      cout<<"out";
      break;
    case vdd:
      cout<<"vdd";
      break;
    case gnd:
      cout<<"gnd";
      break;
    default:
      cout<<"Undefined";
    } //switch
} //Print_pin_type

//Model of a pin/pad pair

//public functions
Pin_class::Pin_class()
{
  Clear(); //sets signal values
  configured = false;
  //initialize
  assertion_level = asserted_high;
  type = dedicated_input;
  direction = input;
} //constructor


Pin_class::~Pin_class()
{
} //destructor


void Pin_class::Configure(pin_type new_type,
			  assertion_type new_assertion_level)
{
  if(configured == false)
    {
      type = new_type;
      assertion_level = new_assertion_level;
      configured = true;
      if((type == vdd)||(type == gnd)||(type == dedicated_input))
	{
	  direction = input;
	} //if
      else
	direction = output;
    } //if
  else
    cerr<<"Error: Can only configure a pin once\n";
} //Configure


void Pin_class::Clear()
{
  internal_value = Unknown;
  pad_value = Floating;
} //Clear


void Pin_class::Update_Pad_Value(voltage_type voltage)
{
  if(direction == input)
    {
      pad_value = voltage;
    }
  else
    {
      cerr<<"Error: This is an output pin"<<endl;
    } //else
} //Update


void Pin_class::Update_Internal_Value(signal_type signal)
{
  if(direction == output)
    {
      internal_value = signal;
    }
  else
    {
      cerr<<"Error: This is an input pin"<<endl;
    } //else
} //Update


void Pin_class::Update()
{
  if(direction == input)
    {
      if(assertion_level == asserted_high)
	internal_value = asserted_high_signal_value(pad_value);
      else
	internal_value = asserted_low_signal_value(pad_value);
    }
  else
    {
      if(assertion_level == asserted_high)
	pad_value = asserted_high_voltage_value(internal_value);
      else
	pad_value = asserted_low_voltage_value(internal_value);
    } //else
} //Update


void Pin_class::Update_Direction(pin_direction_type new_direction)
{
  if((type != io)&&(new_direction != direction))
    {
      cerr<<"\nError: Cannot change pin direction"<<"\n";
    } //if
  else
    {
      direction = new_direction;
    } //else
} //Update signals and direction (if bidirectional)


voltage_type Pin_class::Get_Pad_Value()
{
  return pad_value;
} //Get_Pad_Value


signal_type Pin_class::Get_Internal_Value()
{
  return internal_value;
} //Get_Internal_Value


signal_type Pin_class::Get_Inverted_Internal_Value()
{
  return INV(internal_value);
} //Get_Inverted_Internal_Value


pin_direction_type Pin_class::Get_Direction()
{
  return direction;
} //Get_Direction


pin_type Pin_class::Get_Type()
{
  return type;
} //Get_Type

///////////////////////////////////////////////////////////////////////////////
//end of implementation file





