//cell.h: A header file defining useful cells for CROWBAR
//logic gates, latches

#if !defined(_Cell_h_)
#define _Cell_h_
#include "signal.h"
#include "primitives.h"
#include "custom.h"

//Constants
#define LUT8_Config_Bits 8
#define Configurable_Mux4_Config_Bits 2
#define Configurable_Mux8_Config_Bits 3
#define Receiver_Config_Bits 3

//Combinational logic
////////////////////////////////////////////////////////////////

signal_type TMUX4(signal_type* inputs, signal_type* selects);
//4-way mux with three-state inverters (for CLB)
signal_type TMUX8(signal_type* inputs, signal_type* selects); //for CSB
signal_type GMUX2(signal_type inp0, signal_type inp1, signal_type select1);
//2-way mux with gates (scan-mux, config-mux)
signal_type Pseudo_NMOS_MUX8(signal_type* inputs, signal_type* selects);
//8-way mux used for look-up tables

//Latches: simple memory elements (conditional, resettable)
////////////////////////////////////////////////////////////////

class conditional_latch : public latch
{
//inherits latch class characteristics
//Update function is different (overrides base class implementation)
	public:
		//public functions
		conditional_latch();
		~conditional_latch();
		void Update(signal_type clock, signal_type input, signal_type control);
		//Get_Output is inherited
		//Clear is inherited
	protected:
		//nada
	private:
		//nada	
}; //conditional_latch


class resettable_latch : public latch
{
//inherits latch class characteristics
//Update function is different (overrides base class implementation)
	public:
		//public functions
		resettable_latch();
		~resettable_latch();
		void Update(signal_type clock, signal_type input, signal_type reset);
		//Get_Output is inherited
		//Clear is inherited
	protected:
		//nada
	private:
		//nada	
}; //resettable_latch

//Flops: compound memory elements
////////////////////////////////////////////////////////////////

class config_flop
{
//a-phase vanilla latch bolted onto b-phase conditional latch
	public:
		//public functions
		config_flop();
		~config_flop();
		void Update(signal_type clk_a, signal_type clk_b,
				signal_type enconf, signal_type input);
		signal_type Get_Output(); //Get b_latch value
		void Clear();
	private:
		//private
		latch a_latch;
		conditional_latch b_latch;
}; //config_flop


class resettable_flop
{
//a-phase vanilla latch bolted onto b-phase resettable latch
	public:
		//public functions
		resettable_flop();
		~resettable_flop();
		void Update(signal_type clk_a, signal_type clk_b,
				signal_type reset, signal_type input);
		signal_type Get_Output(); //Get b_latch value
		void Clear();
	private:
		//private
		latch a_latch;
		resettable_latch b_latch;
}; //resettable_flop


//Compound/Combined structures (now to do some real damage)
///////////////////////////////////////////////////////////////

class LUT8
{
 //8 entry lookup table
 //8-way mux with 8-bit shift register
 public:
		//public functions
		LUT8();
		~LUT8();
		void Update(signal_type clk_a, signal_type clk_b,
				signal_type en_conf, signal_type shift_in);
		signal_type Get_Output(signal_type* index);
		signal_type Get_Shift_Output();
		signal_type Get_Config_Bit(unsigned int index);
		void Clear();
 private:
		//private data structures
		config_flop LUT_entries[LUT8_Config_Bits]; //8-bit shift register
}; //LUT8


class configurable_mux4
{
  //4-way mux with config flops
 public:
  configurable_mux4();
  ~configurable_mux4();
  void Update(signal_type clk_a, signal_type clk_b,
	      signal_type en_conf, signal_type shift_in);
  signal_type Get_Output(signal_type* inputs);
  signal_type Get_Shift_Output();
  signal_type Get_Config_Bit(unsigned int index);
  void Clear();
 private:
  //private structures
  config_flop config[Configurable_Mux4_Config_Bits]; //2-bit shift register
}; //input mux


class configurable_mux8
{
  //8-way mux with config flops
 public:
  configurable_mux8();
  ~configurable_mux8();
  void Update(signal_type clk_a, signal_type clk_b,
	      signal_type en_conf, signal_type shift_in);
  signal_type Get_Output(signal_type* inputs);
  signal_type Get_Shift_Output();
  signal_type Get_Config_Bit(unsigned int index);
  void Clear();
 private:
  //private structures
  config_flop config[Configurable_Mux8_Config_Bits]; //3-bit shift register
}; //output mux


class receiver
{
 //Diff-amp, Voltage reference, detector, and 3-bit shift
 //register. Also has staging latches for detector threshold config.
public:
		receiver();
		~receiver();
		void Update(signal_type clk_a, signal_type clk_b,
				signal_type en_conf, signal_type shift_in);
		signal_type Get_Output(optical_type optical_input);
		signal_type Get_Shift_Output();
		signal_type Get_Config_Bit(unsigned int index);
		void Clear();
private:
		//private data structures
		config_flop config[Receiver_Config_Bits]; //shift register
		latch staging_latches[Receiver_Config_Bits]; //see doc
}; //optical receiver

///////////////////////////////////////////////////////////////
//End of header file
#endif


